########################################################################################################
# edex_web_api_request.py
# Author: Troy Kachor
# troyk@phys.ucalgary.ca
#
# Description:
# This script facilitates the downloading of scientific data products from the eDEx API
# 1 - Fetches a list of available files based on specified parameters (e.g., product type, date range, etc.).
# 2 - Displays the list of files and prompts the user to confirm the download.
# 3 - Initiates a download of the selected files, saving them as a single ZIP file (data_products.zip).
# 4 - Tracks and displays the download progress in real-time using a spinner animation.
#
# The script uses multithreading to poll the download progress while the download is in progress.
#
# Usage:
# Modify the PARAMS dictionary to specify the desired data products and filter constraints.
# Run the script and confirm the download when prompted.
#
########################################################################################################

import itertools
import sys
import threading
import time
import uuid

import requests

from app import app

API_URL = 'https://api.edex.phys.ucalgary.ca'

def poll_progress(session_id):
  """
  Polls the /progress endpoint to track the download progress.
  Updates the progress on the same line with a spinner.
  This function is not necessary, but is useful if you wish to track the download progress.
  """
  spinner = itertools.cycle(['-', '/', '|', '\\'])  # Spinner animation
  while True:
    try:
      # Fetch progress data
      progress_response = requests.get(f'{API_URL}/progress?session_id={session_id}')
      progress_data = progress_response.json()

      # Print progress on the same line
      sys.stdout.write(
        f'\rProgress: {progress_data["progress"]:.2f}% '
        f'({progress_data["bytes_sent"]} / {progress_data["total_size"]} bytes) '
        f'{next(spinner)}'
      )
      sys.stdout.flush()  # Ensure the output is displayed immediately

      # Check if the download is complete
      if progress_data['complete']:
        print('\nDownload complete!')
        break

    except requests.exceptions.RequestException as e:
      print(f'\nError fetching progress: {e}')
      break

    # Poll every second
    time.sleep(1)


########################################## Request File list ###########################################
FETCH_DATA_ENDPOINT = f'{API_URL}/api_fetch_data'

"""
  PARAMS is a dictionary to pass to the API as a JSON object 
  Modify this with your desired products and constraints. 
  View additional parameters at the end of the file. 
"""
PARAMS = {
  'product': ['epop-quicklook', 'mgf-quicklook', 'rri-quicklook', 'fai-summary'],
  'date': [['2020-01-01', '2020-01-01']],
  'gap': ['has-data'],
  'rri': ['has-data'],
  'email': 'troyk@phys.ucalgary.ca',
  'query-logic': ' ( 1 AND 2 AND 3 ) ',
}

# Fetch the list of files from the API
req = requests.get(url=FETCH_DATA_ENDPOINT, json=PARAMS)
results = req.json()

for error in results['errors']:
  print(f'eDEx Error: {error}')
for x in results['files']:
  print(f'file: {x}')

"""
 Confirm these are the files you wish to download and press enter to continue.
 You may skip this prompt, but if you are downloading a large amount of data, it is recommended to check
"""
input("Press Enter to download files or ctrl-c to escape...")

########################################### Direct Download ###########################################
DOWNLOAD_ENDPOINT = f'{API_URL}/download'

# Extract filenames from the nested structure
filenames = [file[0] for file in results['files']] 

# Prepare the payload for the download endpoint
session_id = str(uuid.uuid4())  # Generate a unique session ID
download_payload = {
  'selectedFiles': filenames,  # Send only the filenames
  'session_id': session_id,  # Include the session_id, this is to keep track of the download progress
}

# Start the download
print('Starting download...')
try:
  download_response = requests.post(url=DOWNLOAD_ENDPOINT, json=download_payload, stream=True, timeout=(30, 30))

  if download_response.status_code == 200:
    # Start polling for progress in a separate thread so the download can continue. 
    progress_thread = threading.Thread(target=poll_progress, args=(session_id,))
    progress_thread.daemon = True # Make it a dameon thread so it exits when the main thread completes. 
    progress_thread.start()

    # Save the downloaded file
    with open('data_products.zip', 'wb') as f:
      for chunk in download_response.iter_content(chunk_size=8192):
        if chunk:  # Filter out keep-alive chunks
          f.write(chunk)

    # Wait for the progress thread to finish
    progress_thread.join()
    print('Download saved as data_products.zip')
  else:
    print(f'Download failed with status code: {download_response.status_code}')
except requests.exceptions.RequestException as e:
  print(f'Download failed: {e}')


"""
# Dictionary example with every option available for picking and choosing PARAMS for /api_prepare_report. 
# result_filenames is the filename_list that is returned from /api_fetch_data

{   'result_filenames': ['CAS_Orbit_GEO_20141020T000000_20141020T235959_1.1.0.sp3.zip', <...> , 'CAS_Orbit_GEO_20141021T000000_20141021T235959_1.1.0.sp3.zip', 'RRIQL_20180912_105039_105243_4.0.1.png],
    'email': 'troyk@phys.ucalgary.ca',
    'date':  [['2020-06-01', '2020-08-01'], ['2020-11-01', '2021-01-01']]
    'product': [                        'cassiope-orbit-ephemeris-sp3',                      # products
                                        'cassiope-attitude-quaternions-and-ypr',
                                        'cassiope-bus-telemetry',
                                        'cassiope-legacy-ephemeris-and-attitude',
                                        'epop-quicklook',
                                        'cer-quicklook',
                                        'cer-tec',
                                        'cer-ground-receiver-data',
                                        'fai-quicklook',
                                        'fai-summary',
                                        'fai-png-images',
                                        'fai-lv1-hdf5-images',
                                        'gap-quicklook',
                                        'gap-lv1',
                                        'gap-rinex-observation',
                                        'irm-quicklook',
                                        'irm-summary',
                                        'irm-surface-sensor-current',
                                        'irm-lv0b',
                                        'mgf-quicklook',
                                        'mgf-summary',
                                        'mgf-residual',
                                        'mgf-1-sps-lv1b-cdf',
                                        'mgf-160-sps-lv1b-cdf',
                                        'nms-lv0b',
                                        'nms-quicklook',
                                        'nms-quicklook-velocity',
                                        'nms-quicklook-position',
                                        'rri-quicklook',
                                        'rri-lv1-hdf5',
                                        'sei-quicklook',
                                        'sei-summary',
                                        'sei-lv0b'],
    'geophysical-parameters': [         ['ae', '0', '2000'],                            #geophysical constraints
                                        ['ap', '0', '35'],
                                        ['dst', '-600', '100'],
                                        ['f10.7', '0', '300'],
                                        ['kp', '0', '9']],
    'imf-parameters': [                 ['absolute-b', '0', '50'],
                                        ['by-gsm', '-50', '50'],
                                        ['bz-gsm', '-50', '50'],
                                        ['flow-pressure', '0', '20'],
                                        ['flow-speed', '0', '900'],
                                        ['proton-density', '0', '100'],
                                        ['proton-temperature', '0', '1500000']],
    'e-pop-quicklook': [                'has-data'],                                    #epop constraints
    'cer': [                            'has-data',
                                        ['receiver-site', 'arecibo],
                                        ['receiver-site', 'cuzco'],
                                        ['receiver-site', 'eureka'],
                                        ['receiver-site', 'jicamara'],
                                        ['receiver-site', 'siefring'],
                                        ['receiver-site', 'pmaldonad'],
                                        ['receiver-site', 'delta-junct'],
                                        ['receiver-site', 'gakon-ak'],
                                        ['receiver-site', 'valdez-ak']],
    'fai': [                            ['camera-mode', 'high-res'],
                                        ['camera-mode', 'med-res'],
                                        ['camera-mode', 'low-res'],
                                        ['camera-source', 'infrared'],
                                        ['camera-source', 'visible'],
                                        'has-data'],
    'gap': [                            'has-data', 
                                        ['observation-rate', '100hz'],
                                        ['observation-rate', '50hz'],
                                        ['observation-rate', '20hz'],
                                        ['observation-rate', '1hz'],
                                        ['receiver-number', '0'],
                                        ['receiver-number', '1'],
                                        ['receiver-number', '2'],
                                        ['receiver-number', '3'],
                                        ['receiver-number', '4'],
                                        ['receiver-number', 'gap-a'],
                                        ['receiver-number', 'gap-o']],
    'irm': [                            'has-data'],
    'mgf': [                            'has-data'],
    'nms': [                            'has-data'],
    'rri': [                            'has-data',
                                        ['antenna', 'monopole'], 
                                        ['antenna', 'dipole'],
                                        ['bandwidth', '30khz'],
                                        ['bandwidth', '5 KHz'],
                                        ['dwell-a', '0', '60'],
                                        ['dwell-b', '0', '60'],
                                        ['frequency','0','18.5], 
                                        ['gain', 'high'],
                                        ['gain', 'medium'],
                                        ['gain', 'low'],
                                        ['output-channels', 'i1-q1-i3-q3'],
                                        ['output-channels', 'i1-i2-i3-i4'],
                                        ['sweep', 'fixed'],
                                        ['sweep', 'linear'],
                                        ['sweep', 'logarithmic']],
    'sei': [                            'has-data'],
    'spacecraft-position': [            ['altitude', '300', '1500'],
                                        ['eclipse', 'sunlit'],
                                        ['eclipse', 'penumbra'],
                                        ['eclipse', 'umbra'],
                                        ['geographic-latitude', '-81','81'],
                                        ['geographic-longitude', '-180', '180'],
                                        ['magnetic-local-time', '0', '24']],
    'spacecraft-attitude': [            ['nadir-pointing', 'true'],
                                        ['nadir-pointing', 'false'],
                                        ['pitch', '-90', '90']
                                        ['roll',  '-180', '-180'], 
                                        ['yaw',  '-180', '-180']], 
    'amateur-radio': [                  'arrl-field-day',                               #planned experiments
                                        'ham-radio']
    'auroral-region': [                 'auroral-tilt',
                                        'auroral-zone']
    'cusp-region': [                    'cusp-region'] 
    'earth-limb-viewing': [             'earth-limb'],
    'gap-orbit-determination': [        'gap-od'],
    'miscellaneous': [                  'rri-noise-survey',
                                        'rri-special-event',
                                        'rri-vlf',
                                        'churchill-vlf',
                                        'lucky-lake-vlf',
                                        'whistler-detection',
                                        'flash',
                                        'kilauea-eruption',
                                        'lucky-lake-vlf-dmsp-satellite',
                                        'rri-noise-survey-dmsp-satellite',
                                        'he+-detection',
                                        'lucky-lake-vlf-clyde-river-superdarn',
                                        'gap-uncompressed',
                                        'lucky-lake-vlf-dsx-conjunction',
                                        'rri-nw-hemisphere',
                                        'bbc-documentary',
                                        'science-campaign'],
    'point-of-interest': [              'san-francisco',
                                        'revelstoke',
                                        'portsmouth',
                                        'hurricane-dorian',
                                        'fort-mcmurray-fire',
                                        'calgary',
                                        'boston',
                                        'hurricane-sandra'],
    'polar-cap-region': [               'polar-cap'],
    'rocket-satellite-conjunctions': [  'dmsp-satellite',
                                        'cygnus-spacecraft',
                                        'cluster-satellite',
                                        'dsx-conjunction',
                                        'ici-4-rocket',
                                        'ici-5-rocket',
                                        'renu2-rocket',
                                        'techdemosat-satellite',
                                        'zacube-1-satellite'],
    'saa-region': [                     'saa-zone',
                                        'sei-saa'],
    'slew-targets': [                   'poker-flat',
                                        'fort-smith',
                                        'ice-bear-radar'],
    'solar-eclipse': [                  'solar Eclipse',
                                        'solar-eclipse-buenos-aires'],
    'swarm': [                          'swarm-a',
                                        'swarm-b',
                                        'swarm-c'],
    'terrestrial-receiver': [           'autumn-x-receiver',
                                        'lsbb-receiver',
                                        'carisma-receiver',
                                        'above-receiver']
    'terrestrial-transmitter':  [       'tiger-superdarn',
                                        'rothr',
                                        'wwv-radio-station',
                                        'wake-island-transmitter',
                                        'toolik-lake-transmitter',
                                        'dho38-radio-station',
                                        'bpm-time-signal-transmitter',
                                        'kharkiv-isr',
                                        'cyprus-transmitter',
                                        'codar',
                                        'clyde-river-superdarn',
                                        'risr-transmitter',
                                        'clyde-river-superdarn-saa-zone',
                                        'chu-radio-station',
                                        'cambridge-bay-cadi',
                                        'arecibo-transmitter',
                                        'surface-wave-radar',
                                        'sura-heater',
                                        'sondrestrom-radar',
                                        'saskatoon-superdarn',
                                        'isr-world-day',
                                        'new-mexico-transmitter',
                                        'nlk-radio-station',
                                        'nwc-radio-station',
                                        'ottawa-nrcan-transmitter',
                                        'millstone-hill-transmitter',
                                        'kodiak-superdarn',
                                        'polardarn',
                                        'prince-george-superdarn',
                                        'rankin-inlet-superdarn',
                                        'naa-transmitter',
                                        'resolute-bay-transmitter',
                                        'istok',
                                        'hawaii-transmitter',
                                        'hankasalmi-superdarn',
                                        'haarp',
                                        'eiscat-tromso',
                                        'eiscat-svalbard',
                                        'drdc']
                   }
"""
